#____________________________________________________________________________________
#
#    PersistentThreads (see README.txt for more information)
#    Copyright (C) 2002 Christoph Becker-Freyseng
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#    christoph@mmc-startup.com
#    http://www.beyond-thoughts.com
#    http://www.mmc-startup.com
#____________________________________________________________________________________
#



__doc__="""PPManager is your Interface to PersistentThreads. With an instance of
	    PPManager your can add, del, start, stop and retrieve status of 
	    a PersistentProcessor.
	    You can get an PPManager-Instance by calling _getPPMan(zObj), where 
	    zObj is a Zope-Object which is part of the Acquisition- and 
	    Persistence-Tree. Or you can add the PPManUtilMixin to your
	    Zope-Classes, so you can just say self._getPPMan 
	    (then your Zope-Object will be used as zObj automagically)"""
__version__='0.2.1'



from PersistentDict import PersistentDict
from PConstants import product_id, persistent_threads_id
import PThreadMan
import PPHandler
import PTTools

from App.Common import aq_base

import threading






class PPManager:
    _lock= threading.Lock()
    
    def __init__(self, prod):
	self.prod= prod
	PPManager._lock.acquire()
	try:
	    if hasattr(PPManager,'_init'):
		PPManager._init(self)
		delattr(PPManager,'_init')
	except:
	    import traceback, sys
	    sys.stderr.write('During Initialization of PersistentThreads the following error occurred:\n')
	    traceback.print_exc(file=sys.stderr)
	PPManager._lock.release()
    
    def _init(self):
	PPHandler.initPPHandler(self.prod)	# Make sure that it exists
	PPManager.threadMan= PThreadMan.PThreadMan()
	ppInfo= self._getPPInfoDict()
	for (ptID, pp) in self._getPPDict().items():
	    (ppUser, ppPasswd, ppWrapIntoPath, ppAlive)= ppInfo[ptID]
	    if ppAlive:
		self.threadMan.addPP(ptID, ppUser, ppPasswd)
		self.threadMan.startPP(ptID)	# Just use normal delay Zope is not really running at this moment.   Here it's OK to start them immediately, cause they are alreay in ZODB
    
    def _getPPDict(self):
	try:
	    ppDict= getattr(self.prod, persistent_threads_id, None)
	except:
	    ppDict= None
##	ppDict= None	# Force a new ppDict (Debugging)
	if ppDict is None:
	    ppDict= PersistentDict()    # not a normal dictionary! Important for Persistence AND Acquisition
	    setattr(self.prod, persistent_threads_id, ppDict)
	    self.prod._p_changed= 1
	    get_transaction().note('Added a "'+str(persistent_threads_id)+'" to the "'+str(self.prod)+'" Product')
	    get_transaction().commit()
	ppDict= ppDict.__of__(self.prod)
	return ppDict
    
    def _getPPInfoDict(self):
	try:
	    ppInfoDict= getattr(self.prod, persistent_threads_id+'_info', None)
	except:
	    ppInfoDict= None
##	ppInfoDict= None	# Force a new ppInfoDict (Debugging)
	if ppInfoDict is None:
	    ppInfoDict= PersistentDict()    # not a normal dictionary! Important for Persistence AND Acquisition
	    setattr(self.prod, persistent_threads_id+'_info', ppInfoDict)
	    self.prod._p_changed= 1
	    get_transaction().note('Added a "'+str(persistent_threads_id+'_info')+'" to the "'+str(self.prod)+'" Product')
	    get_transaction().commit()
	ppInfoDict= ppInfoDict.__of__(self.prod)
	return ppInfoDict
    
    def hasPP(self, ptID):
	return self._getPPDict().has_key(ptID)
    
    def addPP(self, ptID, ppP, ppUser, ppPasswd, ppWrapIntoPath=None):  # return 1 if succesful or 0 if failed
	ptID= str(ptID)	# Only Strings are valid as Identifiers
	dictD= self._getPPDict()
	if dictD.has_key(ptID):
	    return 0
	dictD[ptID]= ppP
	self._getPPInfoDict()[ptID]= [ppUser, ppPasswd, ppWrapIntoPath, 0]	# 0 -> Not alive now
	self.threadMan.addPP(ptID, ppUser, ppPasswd)
	return 1
    
    def getPP(self, ptID):
	return self._getPPDict().get(ptID, None)
    
    def getPPInfo(self, ptID):
	infD= self._getPPInfoDict().get(ptID, None)
	if infD is None:
	    return None
	else:
	    return tuple(infD)
    
    def changePPInfo(self, ptID, ppUser, ppPasswd, ppWrapIntoPath, ppAlive=None):  # return 1 if succesful or 0 if failed
	infoDict= self._getPPInfoDict()
	ppInfoD= infoDict.get(ptID, None)
	if ppInfoD is None:
	    return 0
	else:
	    self.threadMan.changePP(ptID, ppUser, ppPasswd)
	    ppInfoD[0]= ppUser
	    ppInfoD[1]= ppPasswd
	    ppInfoD[2]= ppWrapIntoPath
	    if ppAlive is not None:
		ppInfoD[3]= ppAlive
	    infoDict._p_changed= 1
	    return 1
    
    def manipulatePP(self, ptID, manipulatorP):
	ppD= self._getPPDict().get(ptID, None)
	if ppD is not None:
	    infoDict= self._getPPInfoDict()[ptID]       # (ppUser, ppPasswd, ppWrapIntoPath, ppAlive)
	    return self.threadMan.manipulatePP(ptID, infoDict[0], infoDict[1], manipulatorP)
	else:	# did not exist
	    return None
    
    def delPP(self, ptID):
	dictD= self._getPPDict()
	if dictD.has_key(ptID):
	    del dictD[ptID]
	    del self._getPPInfoDict()[ptID]
	    return 1
	else:
	    return 0
    
    def startPP(self, ptID, delayP=10.3):  # return 1 if succesful or 0 if failed
	ppD= self._getPPDict().get(ptID, None)
	if ppD is not None:
	    ppInfoD= self._getPPInfoDict()[ptID]	# (ppUser, ppPasswd, ppWrapIntoPath, ppAlive)
	    if not ppInfoD[3]:
		self.changePPInfo(ptID, ppInfoD[0], ppInfoD[1], ppInfoD[2], 1)		# Here there are no conflict-problems, because it surely isn't and wasn't running now
	    if not self.threadMan.hasPP(ptID):	# this is a zombie, wake him
		self.threadMan.addPP(ptID, ppInfoD[0], ppInfoD[1])
	    return self.threadMan.startPP(ptID, delayP)
	else:
	    return 0
    
    def stopPP(self, ptID):  # return 1 if succesful or 0 if failed, BLOCKS until pThread is stopped!
	if self.threadMan.hasPP(ptID):
	    if self.threadMan.stopPP(ptID):
		ppInfoD= self._getPPInfoDict()[ptID]
		self.changePPInfo(ptID, ppInfoD[0], ppInfoD[1], ppInfoD[2], 0)
		return 1
	    else:
		return 0
	else:	# did not exist/run
	    return None
    
    def statusPP(self, ptID):	# status: -666 means Zombie, None=does not exist, -1=dying, 0=running, >0 means running with this waitTime
	dictD= self._getPPDict()
	if dictD.has_key(ptID):
	    return self.threadMan.statusPP(ptID)
	else:	# did not exist
	    return None






def _getPPMan(zObj):
    app= PTTools._getAppObj(zObj)
    prodD= getattr(app.Control_Panel.Products, product_id)	# If we can't find our product we have an unresolve-able problem -> crash OK
    return PPManager(prodD)

class PPManUtilMixin:
    _getPPMan= _getPPMan







